#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2018
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#

import os
import struct
import sys
import unittest
import warnings

try:
    import crc32c
    sw_mode = os.environ.get('CRC32C_SW_MODE')
    if sw_mode == 'none' and not crc32c.hardware_based:
        raise RuntimeError('"none" should force hardware support')
    elif sw_mode == 'force' and crc32c.hardware_based:
        raise RuntimeError('"force" should force software support')
except ImportError:
    crc32c = None

def ulonglong_as_bytes(x):
    return struct.pack('<Q', x)
def uint_as_bytes(x):
    return struct.pack('<I', x)
def ushort_as_bytes(x):
    return struct.pack('<H', x)
def uchar_as_bytes(c):
    return struct.pack('<B', c)

if sys.version_info[0] == 2:
    def as_individual_bytes(x):
        return x
else:
    def as_individual_bytes(x):
        for b in x:
            yield bytes([b])

class warning_catcher(object):
    '''Encapsulates proper warning catch-all logic in python 2.7 and 3'''

    def __init__(self):
        self.catcher = warnings.catch_warnings(record=True)

    def __enter__(self):
        ret = self.catcher.__enter__()
        if sys.version_info[0] == 2:
            warnings.simplefilter("always")
        return ret

    def __exit__(self, *args):
        self.catcher.__exit__(*args)


@unittest.skipIf(crc32c is None, 'no crc32c support in this platform')
class TestMisc(unittest.TestCase):

    def test_zero(self):
        self.assertEqual(0, crc32c.crc32c(b''))

    def test_crc32_deprecated(self):
        with warning_catcher() as warns:
            crc32c.crc32(b'')
        self.assertEqual(len(warns), 1)
        with warning_catcher() as warns:
            crc32c.crc32c(b'')
        self.assertEqual(len(warns), 0)


    def test_msvc_examples(self):
        # Examples taken from MSVC's online examples.
        # Values are not xor'd in the examples though, so we do it here
        max32 = 0xffffffff
        def assert_msvc_vals(b, crc, expected_crc):
            self.assertEqual(expected_crc ^ max32, crc32c.crc32c(b, crc ^ max32))
        assert_msvc_vals(uchar_as_bytes(100), 1, 1412925310)
        assert_msvc_vals(ushort_as_bytes(1000), 1, 3870914500)
        assert_msvc_vals(uint_as_bytes(50000), 1, 971731851)
        assert_msvc_vals(ulonglong_as_bytes(0x88889999EEEE3333), 0x5555AAAA, 0x16F57621)


numbers1 = ('Numbers1', b'123456789', 0xe3069283)
numbers2 = ('Numbers2', b'23456789', 0xBFE92A83)
numbers3 = ('Numbers3', b'1234567890', 0xf3dbd4fe)
phrase = ('Phrase', b'The quick brown fox jumps over the lazy dog', 0x22620404)
long_phrase = ('LongPhrase', (b"Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nunc omni virtuti vitium contrario nomine opponitur. "
b"Conferam tecum, quam cuique verso rem subicias; Te ipsum, dignissimum maioribus tuis, voluptasne induxit, ut adolescentulus eriperes "
b"P. Conclusum est enim contra Cyrenaicos satis acute, nihil ad Epicurum. Duo Reges: constructio interrete. Tum Torquatus: Prorsus, inquit, assentior;\n"
b"Quando enim Socrates, qui parens philosophiae iure dici potest, quicquam tale fecit? Sed quid sentiat, non videtis. Haec quo modo conveniant, non "
b"sane intellego. Sed ille, ut dixi, vitiose. Dic in quovis conventu te omnia facere, ne doleas. Quod si ita se habeat, non possit beatam praestare "
b"vitam sapientia. Quis suae urbis conservatorem Codrum, quis Erechthei filias non maxime laudat? Primum divisit ineleganter; Huic mori optimum esse "
b"propter desperationem sapientiae, illi propter spem vivere."), 0xfcb7575a)

class Crc32cChecks(object):
    def test_all(self):
        self.assertEqual(self.checksum, crc32c.crc32c(self.val))
    def test_piece_by_piece(self):
        c = 0
        for x in as_individual_bytes(self.val):
            c = crc32c.crc32c(x, c)
        self.assertEqual(self.checksum, c)

# Generate the actual unittest classes for each of the testing values
if crc32c is not None:
    for name, val, checksum in (numbers1, numbers2, numbers3, phrase, long_phrase):
        classname = 'Test%s' % name
        locals()[classname] = type(classname, (unittest.TestCase, Crc32cChecks), {'val': val, 'checksum': checksum})


#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2014
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import glob
import platform

import distutils.ccompiler
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext


crcmod_ext = Extension('crc32c',
                       define_macros=[('NDEBUG', None)],
                       depends=glob.glob('*.h'),
                       language='c',
                       sources=['_crc32c.c', 'checkarm.c', 'checksse42.c', 'crc32c_adler.c', 'crc32c_arm64.c', 'crc32c_sw.c'],
                       include_dirs=['.'])

def get_extra_compile_args(is_intel, is_arm):
    # msvc is treated specially; otherwise we assume it's a unix compiler
    comp = distutils.ccompiler.get_default_compiler()
    if comp == 'msvc':
        return ['/O2']
    elif is_intel:
        return ['-O3', '-msse4.2', '-mpclmul']
    elif is_arm:
        return ['-O3', '-march=armv8-a+crc+crypto']
    else:
        return ['-O3']

class _build_ext(build_ext):
    """Custom build_ext command that includes extra compilation arguments"""

    user_options = build_ext.user_options + [
        ('platform=', None, 'The target platform name')]

    def initialize_options(self):
        build_ext.initialize_options(self)
        self.platform = platform.machine()

    def run(self):
        assert(len(self.distribution.ext_modules) == 1)
        platform = self.platform.lower()
        is_intel = platform in ['x86_64', 'amd64', 'i386', 'i686']
        is_arm = platform in ['aarch64_be', 'aarch64', 'armv8b', 'armv8l']
        distutils.log.info("platform: %s, is_intel: %d, is_arm: %d", platform, is_intel, is_arm)
        self.distribution.ext_modules[0].extra_compile_args = get_extra_compile_args(is_intel, is_arm)
        if is_intel:
            self.distribution.ext_modules[0].define_macros += [('IS_INTEL', None)]
        elif is_arm:
            self.distribution.ext_modules[0].define_macros += [('IS_ARM', None)]
        build_ext.run(self)


classifiers = [
    # There's no more specific classifier for LGPLv2.1+
    "License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)",
    "Operating System :: OS Independent",
    "Programming Language :: C",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3.2",
    "Programming Language :: Python :: 3.3",
    "Programming Language :: Python :: 3.4",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
]

with open('README.rst', 'rt') as f:
    long_description = f.read()

setup(name='crc32c',
      author='The ICRAR DIA Team',
      url='https://github.com/ICRAR/crc32c',
      author_email='rtobar@icrar.org',
      version='2.2',
      license="LGPLv2.1+",
      description=('A python package implementing the crc32c algorithm'
      'in hardware and software'),
      long_description=long_description,
      long_description_content_type='text/x-rst',
      classifiers=classifiers,
      ext_modules=[crcmod_ext],
      cmdclass = {'build_ext': _build_ext},
      test_suite="test")
